// sndvolcp Copyright (c)2002-2003 Tomoya Tokairin
// license: GPL

#include <windows.h>
#include <stdio.h>

#define VER "sndvolcp v1.0 1/17/2003 (c)2002-2003 Tomoya Tokairin"

#define MAX_LINE 16
#define MAX_LINECONTROLS 16
#define MAX_CHN 10


//---------------------------------------------------------
// get mixer lines
// Return value is number of lines.
DWORD GetMixerLine(
				   // input
				   HMIXER hMixer, 
				   MIXERCAPS mixerCaps, 
				   DWORD dwComponentType, 
				   
				   // output
				   MIXERLINE* lpMixerLine 
				   )
{
	MIXERLINE mixerLine[MAX_LINE],mixerLineTmp;
	DWORD i,i2;

	// get master lines
	for (i = 0;i < mixerCaps.cDestinations; i++)
	{
		memset(&mixerLine[i],0,sizeof(MIXERLINE));
		mixerLine[i].cbStruct = sizeof(MIXERLINE);
		mixerLine[i].dwDestination = i; 

		if(mixerGetLineInfo(
			(HMIXEROBJ)hMixer,
			&mixerLine[i], 
			MIXER_OBJECTF_HMIXER 
			| MIXER_GETLINEINFOF_DESTINATION  // use dwDestination
			) != MMSYSERR_NOERROR) return 0;
	}

#ifdef _DEBUG
	fprintf(stderr,"\nmaster lines:\n");
	for (i = 0;i < mixerCaps.cDestinations; i++)
		fprintf(stderr,"[%2d] %s\n",i,mixerLine[i].szName);
#endif

	// get lines
	for (i = 0;i < mixerCaps.cDestinations; i++)
	{
		if(mixerLine[i].dwComponentType == dwComponentType)
		{
			for (i2 = 0;i2 < mixerLine[i].cConnections; i2++)
			{
				memset(&mixerLineTmp,0,sizeof(MIXERLINE));
				mixerLineTmp.cbStruct = sizeof(MIXERLINE);
				mixerLineTmp.dwDestination = mixerLine[i].dwDestination; 
				mixerLineTmp.dwSource = i2;
				
				if(mixerGetLineInfo(
					(HMIXEROBJ)hMixer,
					&mixerLineTmp, 
					MIXER_OBJECTF_HMIXER 
					| MIXER_GETLINEINFOF_SOURCE   // use dwDestination and dwSource
					) != MMSYSERR_NOERROR) return 0;

				lpMixerLine[1+i2] = mixerLineTmp;
			}

			lpMixerLine[0] = mixerLine[i]; // ID of master line is '0'.

#ifdef _DEBUG
			fprintf(stderr,"\nlines:\n");
			for (i2 = 0;i2 <= mixerLine[i].cConnections; i2++)
				fprintf(stderr,"[%2d] %s\n",i2,lpMixerLine[i2].szName);
#endif
			return mixerLine[i].cConnections+1;
		}
	}

	return 0;
}



//---------------------------------------------------------
// get mixer controls
BOOL GetMixerControl(
					 // input
					 HMIXER hMixer, 
					 MIXERLINE mixerLine, 
					 DWORD dwType1, 
					 DWORD dwType2, 

					 // output
					 MIXERCONTROL* lpMixerCtrl
					 )
{
	MIXERLINECONTROLS mixerLineCtrl;
	MIXERCONTROL mixerCtrl[MAX_LINECONTROLS];

	DWORD i,dwType;

	memset(&mixerLineCtrl,0,sizeof(MIXERLINECONTROLS));
	mixerLineCtrl.cbStruct = sizeof(MIXERLINECONTROLS);
	mixerLineCtrl.dwLineID = mixerLine.dwLineID;
	mixerLineCtrl.cControls = mixerLine.cControls;
	mixerLineCtrl.cbmxctrl = sizeof(MIXERCONTROL);
	mixerLineCtrl.pamxctrl = mixerCtrl;

	if(mixerGetLineControls(
		(HMIXEROBJ)hMixer,
		&mixerLineCtrl, 
		MIXER_OBJECTF_HMIXER | MIXER_GETLINECONTROLSF_ALL // get all controls
		)
		!= MMSYSERR_NOERROR) return FALSE;

	for(i = 0; i < mixerLineCtrl.cControls; i++ )
	{
		dwType = mixerCtrl[i].dwControlType;
		
		if(dwType == dwType1 || dwType == dwType2)
		{
			*lpMixerCtrl = mixerCtrl[i];
#ifdef _DEBUG
			fprintf(stderr,"\nmixer control: %s\n",mixerLine.szName);
			fprintf(stderr,"control:\n");
			fprintf(stderr,"[%2d] %s\n",i,mixerCtrl[i].szName);
#endif	
			return TRUE;
		}
	}

	return FALSE;
}



//---------------------------------------------------------
// get current status of mute 
BOOL GetStatusOfMute(
			   // input
			   HMIXER hMixer, 
			   MIXERCONTROL mixerCtrl, 
			   
			   // output
			   MIXERCONTROLDETAILS* lpMixerCtrlDetails, 
			   MIXERCONTROLDETAILS_BOOLEAN* mixerCtrlDetBool 
			   )
{
	MIXERCONTROLDETAILS	mixerCtrlDetails;

	memset(&mixerCtrlDetails,0,sizeof(MIXERCONTROLDETAILS));
	mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
	mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
	mixerCtrlDetails.cChannels = 1;
	mixerCtrlDetails.cMultipleItems	= mixerCtrl.cMultipleItems;
	mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_BOOLEAN);
	mixerCtrlDetails.paDetails = mixerCtrlDetBool;
	
	if(mixerGetControlDetails((HMIXEROBJ)hMixer,
		&mixerCtrlDetails,
		MIXER_OBJECTF_HMIXER 
		| MIXER_GETCONTROLDETAILSF_VALUE
		) != MMSYSERR_NOERROR) return FALSE;
	
	*lpMixerCtrlDetails = mixerCtrlDetails;

	return TRUE;
}


//----------------------------------------------------------------------
// get details and volume of the control
BOOL GetVolume(
			   // input
			   HMIXER hMixer, 
			   MIXERLINE mixerLine, 
			   MIXERCONTROL mixerCtrl, 

			   // output
			   MIXERCONTROLDETAILS* lpMixerCtrlDetails, 
			   DWORD dwVolume[MAX_CHN] 
			   )
{
	MIXERCONTROLDETAILS	mixerCtrlDetails;

	memset(&mixerCtrlDetails,0,sizeof(MIXERCONTROLDETAILS));
	mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
	mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
	mixerCtrlDetails.cChannels = mixerLine.cChannels;
	mixerCtrlDetails.cMultipleItems	= mixerCtrl.cMultipleItems;
	mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
	mixerCtrlDetails.paDetails = dwVolume;
	
	if(mixerGetControlDetails((HMIXEROBJ)hMixer,
		&mixerCtrlDetails,
		MIXER_OBJECTF_HMIXER 
		| MIXER_GETCONTROLDETAILSF_VALUE
		) != MMSYSERR_NOERROR) return FALSE;
	
	*lpMixerCtrlDetails = mixerCtrlDetails;
	
	return TRUE;
}



//---------------------------------------------------------
// main
int main(int argc, char* argv[]){

	DWORD dwMixerID,dwMixNum,dwControl;
	HMIXER hMixer = NULL;
	MIXERCAPS mixerCaps;

	// lines
	DWORD dwLineNum;
	MIXERLINE mixerLine[MAX_LINE]; 
	BOOL bEnableLine[MAX_LINE];

	// mute and volume controls
	MIXERCONTROL mixerMuteCtrl[MAX_LINE]; 
	MIXERCONTROL mixerVolCtrl[MAX_LINE]; 

	// details of mute control
	MIXERCONTROLDETAILS	mixerCtrlDetails_mute; 
	MIXERCONTROLDETAILS_BOOLEAN mixerCtrlDetBool_mute;
	
	// details of volume control
	MIXERCONTROLDETAILS mixerCtrlDetails_vol;
	DWORD dwVolume[MAX_CHN];

	DWORD i,i2,dwFoo;
	char szErr[256];

	szErr[0] = '\0';
	fprintf(stderr,"%s\n\n",VER);
	if(argc == 1)
	{
		fprintf(stdout,"sndvolcp id [control] [o/x] [vol1] [vol2] ...\n\n");
	}

	memset(bEnableLine,0,sizeof(BOOL)*MAX_LINE);

	//--------------------------------------------------

	dwMixNum = mixerGetNumDevs();
    if(dwMixNum == 0) goto L_ERR;

	fprintf(stderr,"mixers:\n");
	for(i=0;i < dwMixNum; i++){
		if(mixerGetDevCaps(i,&mixerCaps,sizeof(MIXERCAPS)) != MMSYSERR_NOERROR){
			strcpy(szErr,"mixerGetDevCaps");
			goto L_ERR;
		}
		fprintf(stderr,"[%2d] %s\n",i,mixerCaps.szPname);
	}

	if(argc == 1) goto L_EXIT;

	// get devcaps
	dwMixerID = atoi(argv[1]);
	if(dwMixerID >= dwMixNum) dwMixerID = 0;
	if(mixerGetDevCaps(dwMixerID,&mixerCaps,sizeof(MIXERCAPS)) != MMSYSERR_NOERROR){
		strcpy(szErr,"mixerGetDevCaps");
		goto L_ERR;
	}

	// open mixer
    if(mixerOpen(&hMixer,dwMixerID,0,0,MIXER_OBJECTF_MIXER) != MMSYSERR_NOERROR)
	{
		strcpy(szErr,"mixerOpen");
		goto L_ERR;
	}
	fprintf(stderr,"\nopen: %s\n",mixerCaps.szPname);

	// get mixer lines
	dwLineNum = GetMixerLine(hMixer,mixerCaps,
		MIXERLINE_COMPONENTTYPE_DST_SPEAKERS,

		mixerLine);

	if(dwLineNum == 0)
	{
		strcpy(szErr,"GetMixerLine");
		goto L_ERR;
	}

	for(i=0;i<dwLineNum;i++)
	{
		// get mixer volume controls
		if(!GetMixerControl(hMixer,
			mixerLine[i],
			MIXERCONTROL_CONTROLTYPE_VOLUME,MIXERCONTROL_CONTROLTYPE_MIXER,

			&mixerVolCtrl[i])) continue;

		// get mixer mute controls
		if(!GetMixerControl(hMixer,
			mixerLine[i],
			MIXERCONTROL_CONTROLTYPE_MUTE,MIXERCONTROL_CONTROLTYPE_MUTE,

			&mixerMuteCtrl[i])) continue;

		bEnableLine[i] = TRUE;
	}

	if(argc >= 4)
	{
		// select control
		dwControl = min((DWORD)atoi(argv[2]),dwLineNum);
		fprintf(stderr,"\n[%s]\n",mixerLine[dwControl].szName);

		// mute
		
		// get current status of mute
		if(!GetStatusOfMute(hMixer,mixerMuteCtrl[dwControl],
			&mixerCtrlDetails_mute,&mixerCtrlDetBool_mute))
		{
			strcpy(szErr,"GetStatusOfMute");
			goto L_ERR;
		}
		
		// mute
		fprintf(stderr,"mute: ");
		if(strcmp(argv[3],"x")==0)
		{
			mixerCtrlDetBool_mute.fValue = TRUE;
			fprintf(stderr,"on\n");
		}
		else
		{
			mixerCtrlDetBool_mute.fValue = FALSE;
			fprintf(stderr,"off\n");
		}			
		
		if(mixerSetControlDetails((HMIXEROBJ)hMixer,
			&mixerCtrlDetails_mute,
			MIXER_OBJECTF_HMIXER 
			| MIXER_SETCONTROLDETAILSF_VALUE 
			) != MMSYSERR_NOERROR)
		{
			strcpy(szErr,"mixerSetControlDetails");
			return FALSE;
		}

		// change volume
		if(argc >= 5)
		{
			// get current volume
			if(!GetVolume(hMixer,mixerLine[dwControl],mixerVolCtrl[dwControl],

				&mixerCtrlDetails_vol,dwVolume))
			{
				strcpy(szErr,"GetVolume");
				goto L_ERR;
			}
			
			// set volume
			for(i2=0;i2<mixerCtrlDetails_vol.cChannels;i2++)
			{
				if(4+i2 < (DWORD)argc) dwFoo = (DWORD)atoi(argv[4+i2]);
				
				dwFoo = min(dwFoo,mixerVolCtrl[dwControl].Bounds.dwMaximum);
				fprintf(stderr,"%s: %d -> %d\n",
					"L\0R"+2*i2,
					dwVolume[i2],
					dwFoo);
				
				dwVolume[i2] = dwFoo;
			}
			
			if(mixerSetControlDetails((HMIXEROBJ)hMixer,
				&mixerCtrlDetails_vol,
				MIXER_OBJECTF_HMIXER 
				| MIXER_SETCONTROLDETAILSF_VALUE 
				) != MMSYSERR_NOERROR)
			{
				strcpy(szErr,"mixerSetControlDetails");
				goto L_ERR;
			}
		}
	}

	// show current status
	fprintf(stderr,"\nvolume [x: mute]:\n",i);
	
	for(i=0;i<dwLineNum;i++)
	{
		if(bEnableLine[i])
		{
			
			// get status of mute
			if(!GetStatusOfMute(hMixer,mixerMuteCtrl[i],
				
				&mixerCtrlDetails_mute,&mixerCtrlDetBool_mute))
			{
				strcpy(szErr,"GetStatusOfMute");
				goto L_ERR;
			}
			
			// get volume of each control
			if(!GetVolume(hMixer,mixerLine[i],mixerVolCtrl[i],
				
				&mixerCtrlDetails_vol,dwVolume))
			{
				strcpy(szErr,"GetVolume");
				goto L_ERR;
			}
			
			fprintf(stderr,"[%2d] ",i);
			if(mixerCtrlDetBool_mute.fValue) fprintf(stderr,"x ");
			else  fprintf(stderr,"  ");
			for(i2=0;i2<mixerCtrlDetails_vol.cChannels;i2++)
			{
				fprintf(stderr,"%5d ",dwVolume[i2]);
			}
			fprintf(stderr,"%s\n",mixerLine[i].szName);
		}
	}
	
L_EXIT:

    if(hMixer) mixerClose(hMixer);

	return 0;

L_ERR:

    if(hMixer) mixerClose(hMixer);
	fprintf(stderr,"error in %s\n",szErr);

	return 1;
}


// EOF
